"use strict";
var __defProp = Object.defineProperty;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __publicField = (obj, key, value) => {
  __defNormalProp(obj, typeof key !== "symbol" ? key + "" : key, value);
  return value;
};
const cac = require("cac");
const pc = require("picocolors");
const readline = require("readline");
const vite = require("vite");
const unconfig = require("unconfig");
const fs = require("fs");
const path = require("path");
require("node:url");
require("node:fs");
const pathe = require("pathe");
const fg = require("fast-glob");
const acorn = require("acorn");
const MagicString = require("magic-string");
const micromatch = require("micromatch");
const ws = require("ws");
const zod = require("zod");
const chokidar = require("chokidar");
const EventEmitter = require("events");
const prompt = require("prompts");
const prefix = "[viteburner]";
function formatNormal(first = "", second = "", third = "") {
  const parts = [];
  first && parts.push(pc.green(first));
  second && parts.push(pc.dim(second));
  third && parts.push(third);
  return parts.join(" ");
}
function formatWarn(msg) {
  return pc.yellow(`${msg}`);
}
function formatError(msg) {
  return pc.red(`${msg}`);
}
function createLogger() {
  const logger2 = {
    base: vite.createLogger("info", { prefix, allowClearScreen: false }),
    info: (...msg) => logger2.base.info(formatNormal(...msg), { timestamp: true, clear: false }),
    warn: (...msg) => logger2.base.warn(formatWarn(msg.join(" ")), { timestamp: true, clear: false }),
    error: (...msg) => logger2.base.error(formatError(msg.join(" ")), { timestamp: true, clear: false })
  };
  return logger2;
}
const logger = createLogger();
function onKeypress(handler2) {
  if (!process.stdin.isTTY) {
    logger.warn("current stdin is not a TTY. Keypress events may not work.");
  }
  let running = false;
  async function callback(str, key) {
    if (str === "" || str === "\x1B" || key && key.ctrl && key.name === "c") {
      logger.info("sigterm");
      process.exit(1);
    }
    if (running) {
      return;
    }
    running = true;
    await handler2({ str, key, on, off });
    running = false;
  }
  let rl;
  function on() {
    off();
    rl = readline.createInterface({ input: process.stdin });
    readline.emitKeypressEvents(process.stdin, rl);
    if (process.stdin.isTTY) {
      process.stdin.setRawMode(true);
    }
    process.stdin.on("keypress", callback);
  }
  function off() {
    rl == null ? void 0 : rl.close();
    rl = void 0;
    process.stdin.off("keypress", callback);
    if (process.stdin.isTTY) {
      process.stdin.setRawMode(false);
    }
  }
  on();
  return {
    on,
    off,
    running: () => running
  };
}
let handler;
function setHandler(value) {
  handler = value;
}
onKeypress((ctx) => {
  handler == null ? void 0 : handler(ctx);
});
const name = "viteburner";
const version = "0.5.3";
const description = "Daemon tools of bitburner using vite for script transform, file syncing, RAM monitoring and more!";
const bin = {
  viteburner: "bin/viteburner.js"
};
const main$1 = "./dist/entry.js";
const module$1 = "./dist/entry.mjs";
const exports$1 = {
  ".": {
    require: "./dist/index.js",
    "import": "./dist/index.mjs"
  }
};
const typings = "dist/typings/index.d.ts";
const scripts = {
  build: "vite build",
  test: "npm run lint && npm run test:spec",
  "test:spec": "vitest run",
  "test:playground": "vite-node ./src/entry.ts -- --cwd ./playground",
  postversion: "npm run build",
  lint: "npm run lint:ts && npm run lint:eslint && npm run lint:prettier",
  "lint:ts": "tsc --noEmit",
  "lint:eslint": "eslint --ext .ts,.tsx,.js,.jsx .",
  "lint:prettier": 'prettier --check "**/*.{js,mjs,cjs,jsx,ts,mts,cts,tsx,json}"',
  format: 'prettier --write "**/*.{js,mjs,cjs,jsx,ts,mts,cts,tsx,json}"'
};
const repository = {
  type: "git",
  url: "git+https://github.com/Tanimodori/viteburner.git"
};
const keywords = [
  "bitburner",
  "vite"
];
const author = "Tanimodori";
const license = "MIT";
const bugs = {
  url: "https://github.com/Tanimodori/viteburner/issues"
};
const homepage = "https://github.com/Tanimodori/viteburner#readme";
const devDependencies = {
  "@types/estree": "^1.0.0",
  "@types/micromatch": "^4.0.2",
  "@types/node": "^18.14.1",
  "@types/prompts": "^2.4.2",
  "@types/ws": "^8.5.4",
  "@typescript-eslint/eslint-plugin": "^5.53.0",
  "@typescript-eslint/parser": "^5.53.0",
  eslint: "^8.34.0",
  "eslint-config-prettier": "^8.6.0",
  "eslint-plugin-prettier": "^4.2.1",
  prettier: "^2.8.4",
  typescript: "^4.9.5",
  vite: "^4.1.4",
  "vite-node": "^0.28.5",
  "vite-plugin-dts": "^2.0.2",
  vitest: "^0.27.0"
};
const dependencies = {
  acorn: "^8.8.2",
  cac: "^6.7.14",
  chokidar: "^3.5.3",
  "fast-glob": "^3.2.12",
  "magic-string": "^0.30.0",
  micromatch: "^4.0.5",
  pathe: "^1.1.0",
  picocolors: "^1.0.0",
  prompts: "^2.4.2",
  unconfig: "^0.3.7",
  ws: "^8.12.1",
  zod: "^3.20.6"
};
const peerDependencies = {
  vite: "^3.0.0 || ^4.0.0"
};
const pkg = {
  name,
  version,
  description,
  bin,
  main: main$1,
  module: module$1,
  exports: exports$1,
  typings,
  scripts,
  repository,
  keywords,
  author,
  license,
  bugs,
  homepage,
  devDependencies,
  dependencies,
  peerDependencies
};
function resolveWatchLocation(location) {
  return (filename) => {
    const defaultFilename = defaultUploadLocation(filename);
    let result = location ?? "home";
    if (typeof result === "function") {
      const resolved = result(filename);
      if (!resolved) {
        return [];
      }
      result = resolved;
    }
    if (!Array.isArray(result)) {
      result = [result];
    }
    return result.map((r) => {
      const itemResult = {
        filename: defaultFilename,
        server: "home",
        ...typeof r === "string" ? { server: r } : r
      };
      itemResult.filename = fixStartingSlash(itemResult.filename);
      return itemResult;
    });
  };
}
function resolveDts(dts) {
  if (typeof dts === "string") {
    return dts;
  } else if (dts === false) {
    return void 0;
  } else {
    return "NetscriptDefinitions.d.ts";
  }
}
function resolveDumpFile(dumpFiles) {
  if (typeof dumpFiles === "string") {
    return (file) => dumpFiles + "/" + file;
  } else if (typeof dumpFiles === "function") {
    return dumpFiles;
  } else {
    return void 0;
  }
}
function resolvePolling(usePolling) {
  if (typeof usePolling === "boolean") {
    return { usePolling, pollingOptions: {} };
  } else if (usePolling) {
    return { usePolling: true, pollingOptions: usePolling };
  } else {
    return { usePolling: false, pollingOptions: {} };
  }
}
function resolveConfig(config) {
  var _a, _b, _c, _d;
  const watch = config.watch ?? [];
  const server = ((_a = config.download) == null ? void 0 : _a.server) ?? "home";
  const resolvedConfig = {
    watch: watch.map((item) => ({
      pattern: item.pattern,
      transform: item.transform ?? false,
      location: resolveWatchLocation(item.location)
    })),
    ...resolvePolling(config.usePolling),
    sourcemap: config.sourcemap ?? false,
    port: config.port ?? 12525,
    timeout: config.timeout ?? 1e4,
    dts: resolveDts(config.dts),
    ignoreInitial: config.ignoreInitial ?? false,
    download: {
      server: Array.isArray(server) ? server : [server],
      location: ((_b = config == null ? void 0 : config.download) == null ? void 0 : _b.location) ?? ((file) => "src/" + file),
      ignoreTs: ((_c = config == null ? void 0 : config.download) == null ? void 0 : _c.ignoreTs) ?? true,
      ignoreSourcemap: ((_d = config == null ? void 0 : config.download) == null ? void 0 : _d.ignoreSourcemap) ?? true
    },
    dumpFiles: resolveDumpFile(config.dumpFiles),
    cwd: config.cwd ?? process.cwd()
  };
  return resolvedConfig;
}
function defineConfig(config) {
  return config;
}
async function loadConfig(inlineConfig) {
  const { config } = await unconfig.loadConfig({
    sources: [
      // load from `viteburner.config.xx`
      {
        files: "viteburner.config",
        extensions: ["ts", "mts", "cts", "js", "mjs", "cjs", "json", ""]
      },
      // load inline config from `vite.config`
      {
        files: "vite.config",
        async rewrite(config2) {
          var _a;
          const resolvedConfig = await (typeof config2 === "function" ? config2() : config2);
          const sourcemap = (_a = resolvedConfig == null ? void 0 : resolvedConfig.build) == null ? void 0 : _a.sourcemap;
          return {
            ...sourcemap && { sourcemap },
            ...resolvedConfig == null ? void 0 : resolvedConfig.viteburner
          };
        }
      }
    ],
    cwd: inlineConfig.cwd ?? process.cwd(),
    merge: true
  });
  return resolveConfig({
    ...config,
    ...inlineConfig
  });
}
process.platform === "win32";
function slash(str) {
  return str.replace(/\\/g, "/");
}
function normalizeRequestId(id, base) {
  if (base && id.startsWith(base))
    id = `/${id.slice(base.length)}`;
  return id.replace(/^\/@id\/__x00__/, "\0").replace(/^\/@id\//, "").replace(/^__vite-browser-external:/, "").replace(/^file:/, "").replace(/^\/+/, "/").replace(/\?v=\w+/, "?").replace(/&v=\w+/, "").replace(/\?t=\w+/, "?").replace(/&t=\w+/, "").replace(/\?import/, "?").replace(/&import/, "").replace(/\?&/, "?").replace(/\?+$/, "");
}
class WatchManager {
  constructor(items, options = {}) {
    __publicField(this, "items");
    __publicField(this, "options");
    __publicField(this, "watcher");
    __publicField(this, "initial");
    __publicField(this, "enabled");
    __publicField(this, "enabledTimeStamp");
    __publicField(this, "emitter");
    this.items = items;
    this.options = options;
    this.initial = true;
    this.enabled = true;
    this.enabledTimeStamp = 0;
    this.emitter = new EventEmitter();
  }
  get patterns() {
    return this.items.map((item) => item.pattern);
  }
  findItem(file) {
    return this.items.find((item) => micromatch.isMatch(file, item.pattern));
  }
  init() {
    this.watcher = chokidar.watch(this.patterns, this.options);
    this.watcher.on("ready", () => {
      this.initial = false;
    });
    const events = ["add", "unlink", "change"];
    for (const event of events) {
      this.watcher.on(event, (file) => {
        this.triggerHmr(file, event);
      });
    }
  }
  triggerHmr(file, event) {
    if (!this.enabled) {
      return;
    }
    const root = this.options.cwd ?? process.cwd();
    if (event !== "unlink" && fs.statSync(path.resolve(root, file)).mtimeMs <= this.enabledTimeStamp) {
      return;
    }
    const item = this.findItem(file);
    if (item) {
      this.emitter.emit(hmrPluginName, {
        ...item,
        file: slash(file),
        event,
        initial: this.initial,
        timestamp: Date.now()
      });
    } else {
      throw new Error(`File ${file} does not match any patterns`);
    }
  }
  setEnabled(value) {
    this.enabled = value;
    if (value) {
      this.enabledTimeStamp = Date.now();
    }
  }
  async fullReload() {
    this.enabledTimeStamp = 0;
    const stream = fg.stream(this.patterns, { cwd: this.options.cwd ?? process.cwd() });
    for await (const file of stream) {
      this.triggerHmr(file, "change");
    }
  }
  /** Get all possible filenames to upload */
  getUploadFilenames(filename) {
    filename = removeStartingSlash(slash(filename));
    const item = this.findItem(filename);
    if (!item) {
      return [];
    }
    return item.location(filename);
  }
  /** Shoutcut of `getUploadFilenames(filename).find(server) */
  getUploadFilenamesByServer(filename, server) {
    var _a;
    const filenames = this.getUploadFilenames(filename);
    return (_a = filenames.find((item) => item.server === server)) == null ? void 0 : _a.filename;
  }
  close() {
    var _a;
    (_a = this.watcher) == null ? void 0 : _a.close();
  }
}
function getSourceMapString(map) {
  if (!map)
    return "";
  const mapDataString = JSON.stringify(map);
  const resultA = `//# sourceMappingURL=data:`;
  const resultB = `application/json;base64,${Buffer.from(mapDataString).toString("base64")}`;
  return resultA + resultB;
}
async function writeFile(file, content) {
  const dir = path.dirname(file);
  if (!fs.existsSync(dir)) {
    await fs.promises.mkdir(dir, { recursive: true });
  }
  return fs.promises.writeFile(file, content, {
    flag: "w",
    encoding: "utf8"
  });
}
function isScriptFile(filename) {
  return filename.endsWith(".js") || filename.endsWith(".script");
}
const forceStartingSlash = (s) => {
  return s.startsWith("/") ? s : "/" + s;
};
const fixStartingSlash = (s) => {
  const index = s.lastIndexOf("/");
  if (index === 0) {
    return s.substring(1);
  } else if (index !== -1) {
    return forceStartingSlash(s);
  } else {
    return s;
  }
};
const removeStartingSlash = (s) => {
  return s.startsWith("/") ? s.substring(1) : s;
};
const defaultUploadLocation = (file) => {
  return file.replace(/^src\//, "").replace(/\.ts$/, ".js");
};
const externalRE = /^(https?:)?\/\//;
const isExternalUrl = (url) => externalRE.test(url);
function displayKeyHelpHint() {
  logger.info(
    "help",
    pc.dim("press ") + pc.reset(pc.bold("h")) + pc.dim(" to show help, press ") + pc.reset(pc.bold("q")) + pc.dim(" to exit")
  );
}
function displayWatchAndHelp() {
  logger.info("vite", pc.reset("watching for file changes..."));
  displayKeyHelpHint();
}
function handleKeyInput(wsAdapter) {
  const padding = 18;
  const printStatus = (tag, msg) => {
    logger.info("status", pc.reset(tag.padStart(padding)), msg);
  };
  const displayStatus = () => {
    logger.info("status");
    logger.info("status", " ".repeat(padding - 4) + pc.reset(pc.bold(pc.inverse(pc.green(" STATUS ")))));
    printStatus("connection:", wsAdapter.manager.connected ? pc.green("connected") : pc.yellow("disconnected"));
    printStatus("port:", pc.magenta(wsAdapter.server.config.viteburner.port));
    const pending = wsAdapter.buffers.size;
    const pendingStr = `${pending} file${pending === 1 ? "" : "s"}`;
    const pendingStrStyled = pending ? pc.yellow(pendingStr) : pc.dim(pendingStr);
    printStatus("pending:", pendingStrStyled);
    logger.info("status", pc.dim(""));
  };
  displayStatus();
  displayWatchAndHelp();
  const displayHelp = () => {
    logger.info("help");
    const commands = [
      ["u", "upload all files"],
      ["d", "download all files"],
      ["s", "show status"],
      ["r", "show RAM usage of scripts"],
      ["q", "quit"]
    ];
    logger.info("help", pc.reset(pc.bold("Watch Usage")));
    for (const [key, desc] of commands) {
      logger.info("help", `press ${pc.reset(pc.bold(key))}${pc.dim(" to ")}${desc}`);
    }
    logger.info("help", pc.dim(""));
  };
  const checkConnection = () => {
    if (!wsAdapter.manager.connected) {
      logger.error("conn", pc.red("no connection"));
      return false;
    }
    return true;
  };
  const fullUpload = () => {
    logger.info("upload", pc.reset("force full-upload triggered"));
    wsAdapter.server.watchManager.fullReload();
  };
  const fullDownload = () => {
    logger.info("download", pc.reset("force full-download triggered"));
    wsAdapter.fullDownload();
  };
  const showRamUsageAll = async () => {
    logger.info("ram", pc.reset("fetching ram usage of scripts..."));
    await wsAdapter.getRamUsage();
    return true;
  };
  const showRamUsageGlob = async () => {
    const { pattern } = await prompt({
      type: "text",
      name: "pattern",
      message: "Enter a glob pattern",
      initial: "src/**/*.{ts,js}"
    });
    if (!pattern) {
      return false;
    }
    logger.info("ram", pc.reset("fetching ram usage of scripts..."));
    await wsAdapter.getRamUsage(pattern);
    return true;
  };
  const showRamUsageLocal = async () => {
    const pattern = "**/*.{js,ts,script}";
    const files = await fg(pattern, { cwd: wsAdapter.server.config.root });
    files.sort();
    const fileMap = /* @__PURE__ */ new Map();
    for (const file2 of files) {
      if (file2.endsWith(".d.ts") || file2 === wsAdapter.server.config.viteburner.dts) {
        continue;
      }
      const resolvedData = wsAdapter.getRamUsageLocalData(file2);
      if (resolvedData.length === 0) {
        continue;
      }
      fileMap.set(file2, resolvedData);
    }
    const { file } = await prompt({
      type: "autocomplete",
      name: "file",
      message: "Enter a filename",
      choices: [...fileMap.keys()].map((title) => ({ title }))
    });
    if (!file) {
      return false;
    }
    if (!fs.existsSync(path.resolve(wsAdapter.server.config.root, file))) {
      logger.error("ram", `file ${file} does not exist`);
      return false;
    }
    if (fileMap.has(file)) {
      await wsAdapter.getRamUsageLocalRaw(file, fileMap.get(file));
    } else {
      await wsAdapter.getRamUsageLocal(file);
    }
    return true;
  };
  const showRamUsageRemote = async () => {
    const { server } = await prompt({
      type: "text",
      name: "server",
      message: "Enter a server name",
      initial: "home"
    });
    if (!server) {
      return false;
    }
    const filenames = await wsAdapter.getFileNames(server);
    if (!filenames) {
      return false;
    }
    const { filename } = await prompt({
      type: "autocomplete",
      name: "filename",
      message: "Enter a filename",
      choices: filenames.filter(isScriptFile).map((title) => ({ title }))
    });
    if (!filename) {
      return false;
    }
    await wsAdapter.getRamUsageRemote(server, filename);
    return true;
  };
  const showRamUsageRaw = async () => {
    const { filter } = await prompt({
      type: "select",
      name: "filter",
      message: "Which script do you want to check?",
      initial: 0,
      choices: [
        { title: "All local scripts", value: "all" },
        { title: "Filter local scripts by glob pattern", value: "glob" },
        { title: "Find a local script", value: "local" },
        { title: "Find a remote script", value: "remote" }
      ]
    });
    if (!filter) {
      return false;
    }
    if (filter === "all") {
      return showRamUsageAll();
    } else if (filter === "glob") {
      return showRamUsageGlob();
    } else if (filter === "local") {
      return showRamUsageLocal();
    } else if (filter === "remote") {
      return showRamUsageRemote();
    }
    return false;
  };
  const showRamUsage = async (ctx) => {
    ctx.off();
    const result = await showRamUsageRaw();
    if (result) {
      logger.info("ram", "done");
    } else {
      logger.info("ram", "cancelled");
    }
    ctx.on();
  };
  return async (ctx) => {
    const { key } = ctx;
    let isKeyHandled = true;
    if (key.name === "q") {
      logger.info("bye");
      process.exit();
    } else if (key.name === "s") {
      displayStatus();
    } else if (key.name === "h") {
      displayHelp();
    } else if (key.name === "u") {
      checkConnection() && fullUpload();
    } else if (key.name === "d") {
      checkConnection() && fullDownload();
    } else if (key.name === "r") {
      checkConnection() && await showRamUsage(ctx);
    } else {
      isKeyHandled = false;
    }
    if (isKeyHandled) {
      displayWatchAndHelp();
    }
  };
}
function parse(code) {
  return acorn.parse(code, {
    ecmaVersion: "latest",
    sourceType: "module",
    locations: true
  });
}
function getFilename(options, imports) {
  const realImports = normalizeRequestId(imports ?? options.filename);
  if (isExternalUrl(realImports)) {
    return realImports;
  }
  const importPath = options.manager.getUploadFilenamesByServer(realImports, options.server);
  if (!importPath) {
    const warnInfo = imports ? `Path "${imports}" imported by "${options.filename}" not found on server "${options.server}"` : `File "${imports}" not found on server "${options.server}"`;
    logger.warn(`import`, warnInfo);
    logger.warn(`import`, `This may be a problem when the script is running on the server.`);
    return realImports;
  } else {
    return forceStartingSlash(importPath);
  }
}
function fixImportPath(options) {
  const magicString = new MagicString(options.content);
  const estree = parse(options.content);
  for (const statement of estree.body) {
    if (statement.type === "ImportDeclaration") {
      const source = statement.source;
      const raw = source.raw;
      const value = source.value;
      const importPath = getFilename(options, value);
      const quote = raw[0];
      magicString.overwrite(source.start, source.end, quote + importPath + quote);
    }
  }
  return magicString.toString();
}
const formatUpload = (from, to, serverName) => {
  to = forceStartingSlash(to);
  const dest = `@${serverName}:${to}`;
  return {
    styled: `${pc.dim(from)} ${pc.reset("->")} ${pc.dim(dest)}`,
    raw: `${from} -> ${dest}`
  };
};
const formatDownload = (from, to, serverName) => {
  to = removeStartingSlash(to);
  const src = `@${serverName}:/${from}`;
  return {
    styled: `${pc.dim(src)} ${pc.reset("->")} ${pc.dim(to)}`,
    raw: `${src} -> ${to}`
  };
};
const defaultDownloadLocation = (file) => {
  return "src/" + file;
};
const defaultDts = "NetscriptDefinitions.d.ts";
class WsAdapter {
  constructor(manager, server) {
    __publicField(this, "buffers", /* @__PURE__ */ new Map());
    __publicField(this, "manager");
    __publicField(this, "server");
    this.manager = manager;
    this.server = server;
    this.manager.onConnected(async (ws2) => {
      logger.info("conn", "", "connected");
      const handler2 = () => {
        logger.info("conn", "", pc.yellow("disconnected"));
      };
      ws2.on("close", handler2);
      await this.getDts();
      await this.handleHmrMessage();
      return () => {
        ws2.off("close", handler2);
      };
    });
    this.manager.checkIfWssReused();
  }
  async getDts() {
    const filename = this.server.config.viteburner.dts;
    if (!filename) {
      return;
    }
    try {
      const data = await this.manager.getDefinitionFile();
      const root = this.server.config.root;
      const fullpath = path.resolve(root, filename);
      await writeFile(fullpath, data);
      logger.info("dts change", filename);
    } catch (e) {
      logger.error(`error getting dts file: ${e}`);
    }
  }
  async checkDependencies(data) {
    var _a;
    for (const item of data) {
      if (item.event === "change") {
        continue;
      }
      const resolvedFile = slash(path.resolve(this.server.config.root, item.file));
      (_a = this.server._importGlobMap) == null ? void 0 : _a.forEach((value, key) => {
        if (value.some((pattern) => micromatch.match([resolvedFile], pattern).length > 0)) {
          const importer = slash(path.relative(this.server.config.root, key));
          if (data.some((item2) => item2.file === importer)) {
            return;
          }
          const importerData = this.server.watchManager.findItem(importer);
          if (importerData == null ? void 0 : importerData.transform) {
            data.push({
              file: importer,
              timestamp: item.timestamp,
              initial: item.initial,
              event: "change",
              ...importerData
            });
          }
        }
      });
    }
    return data;
  }
  async handleHmrMessage(data) {
    if (!data) {
      data = [];
    } else if (!Array.isArray(data)) {
      data = [data];
    }
    data = await this.checkDependencies(data);
    const connected = this.manager.connected;
    for (const item of data) {
      this.buffers.set(item.file, item);
      logger.info(`hmr ${item.event}`, item.file, pc.yellow("(pending)"));
    }
    if (!connected) {
      return;
    }
    if (this.buffers.size) {
      for (const item of this.buffers.values()) {
        await this.uploadFile(item);
      }
    }
  }
  deleteCache(data) {
    const currentData = this.buffers.get(data.file);
    if (currentData && data.timestamp === currentData.timestamp) {
      this.buffers.delete(data.file);
    }
  }
  async dumpFile(data, content, server) {
    var _a, _b;
    const relative2 = (_b = (_a = this.server.config.viteburner).dumpFiles) == null ? void 0 : _b.call(_a, data.file, server);
    if (!relative2) {
      return;
    }
    const fullpath = path.resolve(this.server.config.root, relative2);
    await writeFile(fullpath, content);
    logger.info("dump", formatUpload(data.file, slash(relative2), server).styled);
  }
  async fetchModule(data) {
    let content = "";
    if (data.transform) {
      this.server.invalidateFile(data.file);
      const module2 = await this.server.fetchModule(data.file);
      if (!module2) {
        throw new Error("module not found: " + data.file);
      }
      content = module2.code;
      if (this.server.config.viteburner.sourcemap === "inline" && module2.map) {
        content += getSourceMapString(module2.map);
      }
    } else {
      const buffer = await fs.promises.readFile(path.resolve(this.server.config.root, data.file));
      content = buffer.toString();
    }
    return content;
  }
  fixImport(content, data, serverName) {
    if (data.transform) {
      return fixImportPath({
        content,
        filename: data.file,
        server: serverName,
        manager: this.server.watchManager
      });
    } else {
      return content;
    }
  }
  async uploadFile(data) {
    this.deleteCache(data);
    const isAdd = data.event !== "unlink";
    let content = "";
    if (isAdd) {
      try {
        content = await this.fetchModule(data);
      } catch (e) {
        logger.error(String(e));
        return;
      }
    }
    const payloads = this.server.watchManager.getUploadFilenames(data.file);
    if (!payloads.length) {
      logger.info(`hmr ${data.event}`, data.file, pc.dim("(ignored)"));
      return;
    }
    for (const { filename, server: serverName } of payloads) {
      const fileChangeStrs = formatUpload(data.file, filename, serverName);
      try {
        if (isAdd) {
          if (data.transform) {
            content = this.fixImport(content, data, serverName);
          }
          this.dumpFile(data, content, serverName);
          await this.manager.pushFile({
            filename,
            content,
            server: serverName
          });
        } else {
          await this.manager.deleteFile({
            filename,
            server: serverName
          });
        }
        logger.info(`hmr ${data.event}`, fileChangeStrs.styled, pc.green("(done)"));
      } catch (e) {
        logger.error(`error ${data.event}: ${fileChangeStrs.raw} ${e}`);
        logger.error(`hmr ${data.event} ${data.file} (error)`);
        continue;
      }
    }
  }
  async fullDownload() {
    logger.info("vite", pc.reset("stop watching for file changes while downloading"));
    this.server.watchManager.setEnabled(false);
    const servers = this.server.config.viteburner.download.server;
    const filesMap = /* @__PURE__ */ new Map();
    for (const server of servers) {
      try {
        filesMap.set(server, await this.manager.getAllFiles({ server }));
      } catch (e) {
        logger.error(`error: connot get filelist from server ${server}: ${e}`);
        continue;
      }
    }
    for (const [server, files] of filesMap) {
      const { location: locationFn, ignoreTs, ignoreSourcemap } = this.server.config.viteburner.download;
      for (const file of files) {
        file.filename = removeStartingSlash(file.filename);
        const location = locationFn(file.filename, server);
        if (!location) {
          logger.info(`download`, `@${server}:/${file.filename}`, pc.dim("(ignored)"));
          continue;
        }
        const resolvedLocation = path.resolve(this.server.config.root, location);
        const fileChangeStrs = formatDownload(file.filename, location, server);
        try {
          const isIgnoreTs = () => {
            return ignoreTs && resolvedLocation.endsWith(".js") && fs.existsSync(resolvedLocation.substring(0, resolvedLocation.length - 3) + ".ts");
          };
          const isIgnoreSourceMap = () => {
            return ignoreSourcemap && file.content.match(/\/\/# sourceMappingURL=\S+\s*$/g);
          };
          if (isIgnoreTs() || isIgnoreSourceMap()) {
            logger.info(`download`, fileChangeStrs.styled, pc.dim("(ignored)"));
            continue;
          }
          await writeFile(resolvedLocation, file.content);
          logger.info(`download`, fileChangeStrs.styled, pc.green("(done)"));
        } catch (e) {
          logger.error(`download`, fileChangeStrs.raw, `(${e})`);
        }
      }
    }
    logger.info("vite", pc.reset("download completed, watching for file changes..."));
    this.server.watchManager.setEnabled(true);
  }
  async getRamUsage(pattern) {
    const patterns = pattern ?? this.server.watchManager.patterns;
    if (!patterns) {
      logger.warn("ram", "no pattern found");
      return;
    }
    const files = await fg(patterns, { cwd: this.server.config.root });
    if (files.length === 0) {
      logger.warn("ram", "no file found");
      return;
    }
    files.sort();
    for (const file of files) {
      await this.getRamUsageLocal(file);
    }
  }
  getRamUsageLocalData(file) {
    return this.server.watchManager.getUploadFilenames(file);
  }
  async getRamUsageLocalRaw(file, resolvedData) {
    let isScript = false;
    let ramUsage = -1;
    if (resolvedData.length === 0) {
      logger.info("ram", `${file} (ignored)`);
      return true;
    }
    for (const { filename, server } of resolvedData) {
      const formatUploadStrs = formatUpload(file, filename, server);
      if (!isScriptFile(filename)) {
        continue;
      }
      isScript = true;
      try {
        ramUsage = await this.manager.calculateRam({ filename, server });
        logger.info("ram", pc.reset(`${file}: ${ramUsage} GB`));
        break;
      } catch (e) {
        logger.warn(`ram`, formatUploadStrs.raw, `(${e})`);
      }
    }
    if (isScript) {
      if (ramUsage === -1) {
        logger.warn(`ram`, file, `(no target found)`);
        return false;
      }
    } else {
      logger.info("ram", file, pc.dim("(ignored)"));
    }
    return true;
  }
  async getRamUsageLocal(file) {
    const resolvedData = this.getRamUsageLocalData(file);
    return this.getRamUsageLocalRaw(file, resolvedData);
  }
  async getRamUsageRemote(server, filename) {
    const resolvedFilename = fixStartingSlash(filename);
    logger.info("ram", pc.reset("fetching ram usage of scripts..."));
    try {
      const ramUsage = await this.manager.calculateRam({ filename: resolvedFilename, server });
      logger.info("ram", pc.reset(`@${server}/${filename}: ${ramUsage} GB`));
    } catch (e) {
      logger.error(`ram`, `@${server}/${filename}: ${e}`);
    }
  }
  async getFileNames(server) {
    try {
      const filenames = await this.manager.getFileNames({ server });
      return filenames.map(removeStartingSlash);
    } catch (e) {
      logger.error(`list`, `cannot fetch filenames from server ${server}: ${e}`);
      return null;
    }
  }
}
let reused = false;
let wss = null;
let wssPort = -1;
function getWss(port) {
  if (wss === null || wssPort !== port) {
    if (wss) {
      wss.clients.forEach((ws2) => ws2.close());
      wss.close();
    }
    wssPort = port;
    wss = new ws.WebSocketServer({ port });
    reused = false;
  } else {
    reused = true;
  }
  return wss;
}
function isWssReused() {
  return reused;
}
const wsResponseSchema = zod.z.object({
  jsonrpc: zod.z.literal("2.0"),
  id: zod.z.number(),
  result: zod.z.any(),
  error: zod.z.any()
});
const pushFileResponseSchema = zod.z.literal("OK");
const getFileResponseSchema = zod.z.string();
const deleteFileResponseSchema = zod.z.literal("OK");
const getFileNamesResponseSchema = zod.z.array(zod.z.string());
const getAllFilesResponseSchema = zod.z.array(
  zod.z.object({
    filename: zod.z.string(),
    content: zod.z.string()
  })
);
const calculateRamResponseSchema = zod.z.number();
const getDefinitionFileResponseSchema = zod.z.string();
class WsManager {
  constructor(options) {
    __publicField(this, "options");
    __publicField(this, "ws");
    __publicField(this, "wss");
    __publicField(this, "trackers");
    __publicField(this, "nextId");
    __publicField(this, "unregisters");
    this.options = {
      timeout: 1e4,
      ...options
    };
    this.trackers = [];
    this.nextId = 0;
    this.ws = void 0;
    this.wss = getWss(this.options.port);
    this.unregisters = [];
    this._registerHandler();
  }
  get connected() {
    var _a;
    return ((_a = this.ws) == null ? void 0 : _a.readyState) === ws.WebSocket.OPEN;
  }
  _registerHandler() {
    const _onConnected = (ws2) => {
      this.ws = ws2;
      ws2.on("message", (response) => this.handleMessage(response));
    };
    const _onError = (e) => {
      const err = String(e);
      if (err.indexOf("EADDRINUSE") !== -1) {
        logger.error("ws", `fatal: port ${this.options.port} is already in use`);
        process.exit(1);
      } else {
        logger.error("ws", `${err}`);
      }
    };
    this.wss.on("connection", _onConnected);
    this.wss.on("error", _onError);
    this.unregisters.push(() => {
      this.wss.off("connection", _onConnected);
      this.wss.off("error", _onError);
    });
  }
  onConnected(cb) {
    const handler2 = async (ws2) => {
      this.ws = ws2;
      const unregister = await cb(ws2);
      unregister && this.unregisters.push(unregister);
    };
    this.wss.on("connection", handler2);
    this.unregisters.push(() => {
      this.wss.off("connection", handler2);
    });
  }
  checkIfWssReused() {
    if (isWssReused() && this.wss.clients.size > 0) {
      for (const client of this.wss.clients) {
        this.ws = client;
      }
      this.wss.emit("connection", this.ws);
    }
  }
  handleMessage(response) {
    const parsed = wsResponseSchema.parse(JSON.parse(response.toString()));
    const { id, result, error } = parsed;
    if (!this.trackers[id]) {
      return;
    }
    const { resolve, reject } = this.trackers[id];
    if (error) {
      reject(error);
    }
    resolve(result);
  }
  // eslint-disable-next-line @typescript-eslint/no-explicit-any
  async sendMessage(options) {
    const params = options.params;
    if (!this.ws || !this.connected) {
      throw new Error("No connection");
    }
    const id = ++this.nextId;
    this.ws.send(
      JSON.stringify({
        jsonrpc: "2.0",
        id,
        method: options.method,
        ...params && { params }
      })
    );
    const result = new Promise((resolve, reject) => {
      const onResolve = (data) => {
        var _a;
        try {
          resolve(((_a = options.validator) == null ? void 0 : _a.parse(data)) ?? data);
        } catch (e) {
          reject(e);
        } finally {
          delete this.trackers[id];
        }
      };
      const onReject = (reason) => {
        delete this.trackers[id];
        reject(reason);
      };
      this.trackers[id] = {
        resolve: onResolve,
        reject: onReject
      };
    });
    if (this.options.timeout) {
      setTimeout(() => {
        if (this.trackers[id]) {
          this.trackers[id].reject(new Error(`Timeout after ${this.options.timeout}ms`));
        }
      }, this.options.timeout);
    }
    return result;
  }
  async pushFile(params) {
    return this.sendMessage({
      method: "pushFile",
      params,
      validator: pushFileResponseSchema
    });
  }
  async getFile(params) {
    return this.sendMessage({
      method: "getFile",
      params,
      validator: getFileResponseSchema
    });
  }
  async deleteFile(params) {
    return this.sendMessage({
      method: "deleteFile",
      params,
      validator: deleteFileResponseSchema
    });
  }
  async getFileNames(params) {
    return this.sendMessage({
      method: "getFileNames",
      params,
      validator: getFileNamesResponseSchema
    });
  }
  async getAllFiles(params) {
    return this.sendMessage({
      method: "getAllFiles",
      params,
      validator: getAllFilesResponseSchema
    });
  }
  async calculateRam(params) {
    return this.sendMessage({
      method: "calculateRam",
      params,
      validator: calculateRamResponseSchema
    });
  }
  async getDefinitionFile() {
    return this.sendMessage({
      method: "getDefinitionFile",
      validator: getDefinitionFileResponseSchema
    });
  }
  close() {
    this.unregisters.forEach((unregister) => unregister());
  }
}
const hmrPluginName = "viteburner:hmr";
const virtualModuleId = "virtual:viteburner-entry";
function getDefaultConfig() {
  return {
    mode: "development",
    optimizeDeps: { disabled: true },
    clearScreen: false,
    build: {
      lib: {
        /**
         * Meaningless for Vite<3.2.0
         * @see {@link https://github.com/vitejs/vite/discussions/1736}
         */
        entry: virtualModuleId,
        formats: ["es"]
      }
    },
    server: { middlewareMode: true }
  };
}
function viteburnerPlugin(inlineConfig) {
  const resolvedVirtualModuleId = "\0" + virtualModuleId;
  let server;
  let wsAdapter;
  return {
    name: "viteburner",
    apply: "serve",
    // Load viteburner.config.xx, merge with config, and resolve
    async config(config) {
      logger.info("config", "resolving user config...");
      config.viteburner = await loadConfig(inlineConfig);
      return getDefaultConfig();
    },
    configResolved() {
      logger.info("config", "config resolved");
    },
    // save server instance
    configureServer(_server) {
      server = {
        ..._server,
        pathToId(file) {
          const id = `/@fs/${slash(pathe.resolve(server.config.root, file))}`;
          return normalizeRequestId(id, server.config.base);
        },
        async invalidateFile(file) {
          const id = server.pathToId(file);
          const module2 = await server.moduleGraph.getModuleByUrl(id);
          if (module2) {
            server.moduleGraph.invalidateModule(module2);
          }
        },
        async fetchModule(file) {
          const id = server.pathToId(file);
          return server.transformRequest(id);
        },
        onHmrMessage(handler2) {
          server.watchManager.emitter.on(hmrPluginName, (data) => handler2(data, server));
        },
        async buildStart() {
          return server.pluginContainer.buildStart({});
        }
      };
    },
    // main entry
    buildStart() {
      logger.info("watch", "creating a watcher...");
      const { root, viteburner } = server.config;
      const { watch, ignoreInitial, port, timeout, usePolling, pollingOptions } = viteburner;
      server.watchManager = new WatchManager(watch, {
        cwd: root,
        persistent: true,
        ignoreInitial,
        usePolling: !!usePolling,
        ...pollingOptions
      });
      logger.info("ws", "creating ws server...");
      const wsManager = new WsManager({ port, timeout });
      wsAdapter = new WsAdapter(wsManager, server);
      server.onHmrMessage((data) => wsAdapter.handleHmrMessage(data));
      server.watchManager.init();
      setHandler(handleKeyInput(wsAdapter));
    },
    // virtual entry
    resolveId(id) {
      if (id === virtualModuleId) {
        return resolvedVirtualModuleId;
      }
    },
    load(id) {
      if (id === resolvedVirtualModuleId) {
        return `export {}`;
      }
    },
    // exit
    buildEnd() {
      server.watchManager.close();
      wsAdapter.manager.close();
      setHandler();
    }
  };
}
const cli = cac("viteburner");
cli.command("", "start dev server").alias("serve").alias("dev").option("--cwd <cwd>", "Working directory").option("--port <port>", "Port to listen on").action(startDev);
cli.help();
cli.version(pkg.version);
async function startDev(options) {
  const cwd = options.cwd;
  const port = options.port;
  const resolveInlineConfig = {
    ...cwd && { cwd },
    ...port && { port }
  };
  logger.info("version", pkg.version);
  logger.info("vite", "creating dev server...");
  vite.createServer({
    ...cwd && { root: cwd },
    viteburner: resolveInlineConfig,
    plugins: [viteburnerPlugin(resolveInlineConfig)]
  });
}
async function main() {
  cli.parse();
}
exports.WatchManager = WatchManager;
exports.WsAdapter = WsAdapter;
exports.WsManager = WsManager;
exports.calculateRamResponseSchema = calculateRamResponseSchema;
exports.createLogger = createLogger;
exports.defaultDownloadLocation = defaultDownloadLocation;
exports.defaultDts = defaultDts;
exports.defaultUploadLocation = defaultUploadLocation;
exports.defineConfig = defineConfig;
exports.deleteFileResponseSchema = deleteFileResponseSchema;
exports.displayKeyHelpHint = displayKeyHelpHint;
exports.displayWatchAndHelp = displayWatchAndHelp;
exports.externalRE = externalRE;
exports.fixImportPath = fixImportPath;
exports.fixStartingSlash = fixStartingSlash;
exports.forceStartingSlash = forceStartingSlash;
exports.formatDownload = formatDownload;
exports.formatError = formatError;
exports.formatNormal = formatNormal;
exports.formatUpload = formatUpload;
exports.formatWarn = formatWarn;
exports.getAllFilesResponseSchema = getAllFilesResponseSchema;
exports.getDefaultConfig = getDefaultConfig;
exports.getDefinitionFileResponseSchema = getDefinitionFileResponseSchema;
exports.getFileNamesResponseSchema = getFileNamesResponseSchema;
exports.getFileResponseSchema = getFileResponseSchema;
exports.getSourceMapString = getSourceMapString;
exports.getWss = getWss;
exports.handleKeyInput = handleKeyInput;
exports.hmrPluginName = hmrPluginName;
exports.isExternalUrl = isExternalUrl;
exports.isScriptFile = isScriptFile;
exports.isWssReused = isWssReused;
exports.loadConfig = loadConfig;
exports.logger = logger;
exports.main = main;
exports.onKeypress = onKeypress;
exports.prefix = prefix;
exports.pushFileResponseSchema = pushFileResponseSchema;
exports.removeStartingSlash = removeStartingSlash;
exports.resolveConfig = resolveConfig;
exports.resolveDts = resolveDts;
exports.resolveDumpFile = resolveDumpFile;
exports.resolvePolling = resolvePolling;
exports.resolveWatchLocation = resolveWatchLocation;
exports.setHandler = setHandler;
exports.startDev = startDev;
exports.virtualModuleId = virtualModuleId;
exports.viteburnerPlugin = viteburnerPlugin;
exports.writeFile = writeFile;
exports.wsResponseSchema = wsResponseSchema;
